package com.sun.imageio.plugins.bmp;

import javax.imageio.event.IIOWriteProgressListener;
import java.util.Iterator;
import javax.imageio.event.IIOWriteWarningListener;
import javax.imageio.ImageIO;
import java.awt.image.ColorModel;
import java.awt.image.SampleModel;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferByte;
import java.awt.image.MultiPixelPackedSampleModel;
import java.io.IOException;
import java.awt.image.DirectColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.IndexColorModel;
import java.awt.image.BandedSampleModel;
import java.awt.image.SinglePixelPackedSampleModel;
import java.awt.image.ComponentSampleModel;
import com.sun.imageio.plugins.common.ImageUtil;
import java.awt.Rectangle;
import javax.imageio.IIOImage;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.plugins.bmp.BMPImageWriteParam;
import javax.imageio.ImageWriteParam;
import com.sun.imageio.plugins.common.I18N;
import java.nio.ByteOrder;
import javax.imageio.spi.ImageWriterSpi;
import java.io.ByteArrayOutputStream;
import javax.imageio.stream.ImageOutputStream;
import javax.imageio.ImageWriter;

public class BMPImageWriter extends ImageWriter implements BMPConstants
{
    private ImageOutputStream stream;
    private ByteArrayOutputStream embedded_stream;
    private int version;
    private int compressionType;
    private boolean isTopDown;
    private int w;
    private int h;
    private int compImageSize;
    private int[] bitMasks;
    private int[] bitPos;
    private byte[] bpixels;
    private short[] spixels;
    private int[] ipixels;
    
    public BMPImageWriter(ImageWriterSpi originatingProvider) {
        super(originatingProvider);
        this.stream = null;
        this.embedded_stream = null;
        this.compImageSize = 0;
    }
    
    @Override
    public void setOutput(Object output) {
        super.setOutput(output);
        if (output != null) {
            if (!(output instanceof ImageOutputStream)) {
                throw new IllegalArgumentException(I18N.getString("BMPImageWriter0"));
            }
            (this.stream = (ImageOutputStream)output).setByteOrder(ByteOrder.LITTLE_ENDIAN);
        }
        else {
            this.stream = null;
        }
    }
    
    @Override
    public ImageWriteParam getDefaultWriteParam() {
        return new BMPImageWriteParam();
    }
    
    @Override
    public IIOMetadata getDefaultStreamMetadata(ImageWriteParam imageWriteParam) {
        return null;
    }
    
    @Override
    public IIOMetadata getDefaultImageMetadata(ImageTypeSpecifier imageTypeSpecifier, ImageWriteParam imageWriteParam) {
        BMPMetadata bmpMetadata = new BMPMetadata();
        bmpMetadata.bmpVersion = "BMP v. 3.x";
        bmpMetadata.compression = this.getPreferredCompressionType(imageTypeSpecifier);
        if (imageWriteParam != null && imageWriteParam.getCompressionMode() == 2) {
            bmpMetadata.compression = BMPCompressionTypes.getType(imageWriteParam.getCompressionType());
        }
        bmpMetadata.bitsPerPixel = (short)imageTypeSpecifier.getColorModel().getPixelSize();
        return bmpMetadata;
    }
    
    @Override
    public IIOMetadata convertStreamMetadata(IIOMetadata iioMetadata, ImageWriteParam imageWriteParam) {
        return null;
    }
    
    @Override
    public IIOMetadata convertImageMetadata(IIOMetadata iioMetadata, ImageTypeSpecifier imageTypeSpecifier, ImageWriteParam imageWriteParam) {
        return null;
    }
    
    @Override
    public boolean canWriteRasters() {
        return true;
    }
    
    @Override
    public void write(IIOMetadata iioMetadata, IIOImage iioImage, ImageWriteParam defaultWriteParam) throws IOException {
        if (this.stream == null) {
            throw new IllegalStateException(I18N.getString("BMPImageWriter7"));
        }
        if (iioImage == null) {
            throw new IllegalArgumentException(I18N.getString("BMPImageWriter8"));
        }
        this.clearAbortRequest();
        this.processImageStarted(0);
        if (defaultWriteParam == null) {
            defaultWriteParam = this.getDefaultWriteParam();
        }
        BMPImageWriteParam bmpImageWriteParam = (BMPImageWriteParam)defaultWriteParam;
        int n = 24;
        int n2 = 0;
        int mapSize = 0;
        IndexColorModel indexColorModel = null;
        RenderedImage renderedImage = null;
        Raster raster = null;
        int hasRaster = iioImage.hasRaster();
        Rectangle sourceRegion = defaultWriteParam.getSourceRegion();
        this.compImageSize = 0;
        SampleModel sampleModel;
        ColorModel colorModel;
        Rectangle obj;
        if (!(hasRaster == 0)) {
            raster = iioImage.getRaster();
            sampleModel = raster.getSampleModel();
            colorModel = ImageUtil.createColorModel(null, sampleModel);
            if (sourceRegion == null) {
                obj = raster.getBounds();
            }
            else {
                obj = sourceRegion.intersection(raster.getBounds());
            }
        }
        else {
            renderedImage = iioImage.getRenderedImage();
            sampleModel = renderedImage.getSampleModel();
            colorModel = renderedImage.getColorModel();
            Rectangle r = new Rectangle(renderedImage.getMinX(), renderedImage.getMinY(), renderedImage.getWidth(), renderedImage.getHeight());
            if (sourceRegion == null) {
                obj = r;
            }
            else {
                obj = sourceRegion.intersection(r);
            }
        }
        IIOMetadata metadata = iioImage.getMetadata();
        BMPMetadata bmpMetadata;
        if (metadata != null && metadata instanceof BMPMetadata) {
            bmpMetadata = (BMPMetadata)metadata;
        }
        else {
            bmpMetadata = (BMPMetadata)this.getDefaultImageMetadata(new ImageTypeSpecifier(colorModel, sampleModel), defaultWriteParam);
        }
        if (obj.isEmpty()) {
            throw new RuntimeException(I18N.getString("BMPImageWrite0"));
        }
        int sourceXSubsampling = defaultWriteParam.getSourceXSubsampling();
        int sourceYSubsampling = defaultWriteParam.getSourceYSubsampling();
        int subsamplingXOffset = defaultWriteParam.getSubsamplingXOffset();
        int subsamplingYOffset = defaultWriteParam.getSubsamplingYOffset();
        int dataType = sampleModel.getDataType();
        obj.translate(subsamplingXOffset, subsamplingYOffset);
        obj.width -= subsamplingXOffset;
        obj.height -= subsamplingYOffset;
        int x = obj.x / sourceXSubsampling;
        int y = obj.y / sourceYSubsampling;
        this.w = (obj.width + sourceXSubsampling - 1) / sourceXSubsampling;
        this.h = (obj.height + sourceYSubsampling - 1) / sourceYSubsampling;
        int n3 = obj.x % sourceXSubsampling;
        int n4 = obj.y % sourceYSubsampling;
        int equals = new Rectangle(x, y, this.w, this.h).equals(obj);
        int[] sourceBands = defaultWriteParam.getSourceBands();
        boolean n5 = true;
        int n6 = sampleModel.getNumBands();
        if (sourceBands != null) {
            sampleModel = sampleModel.createSubsetSampleModel(sourceBands);
            colorModel = null;
            n5 = false;
            n6 = sampleModel.getNumBands();
        }
        else {
            sourceBands = new int[n6];
            for (int i = 0; i < n6; ++i) {
                sourceBands[i] = i;
            }
        }
        int[] bandOffsets = null;
        int n7 = 1;
        if ((sampleModel instanceof ComponentSampleModel)) {
            bandOffsets = ((ComponentSampleModel)sampleModel).getBandOffsets();
            if ((sampleModel instanceof BandedSampleModel)) {
                n7 = false;
            }
            else {
                for (int k = 0; k < bandOffsets.length; ++k) {
                    n7 &= (bandOffsets[k] == bandOffsets.length - k - 1);
                }
            }
        }
        else if ((sampleModel instanceof SinglePixelPackedSampleModel)) {
            int[] bitOffsets = ((SinglePixelPackedSampleModel)sampleModel).getBitOffsets();
            for (int j = 0; j < bitOffsets.length - 1; ++j) {
                n7 &= (bitOffsets[j] > bitOffsets[j + 1]);
            }
        }
        if (bandOffsets == null) {
            bandOffsets = new int[n6];
            for (int l = 0; l < n6; ++l) {
                bandOffsets[l] = l;
            }
        }
        int n8 = equals & n7;
        int[] sampleSize = sampleModel.getSampleSize();
        int n9 = this.w * n6;
        switch (bmpImageWriteParam.getCompressionMode()) {
            case 2: {
                this.compressionType = BMPCompressionTypes.getType(bmpImageWriteParam.getCompressionType());
                break;
            }
            case 3: {
                this.compressionType = bmpMetadata.compression;
                break;
            }
            case 1: {
                this.compressionType = this.getPreferredCompressionType(colorModel, sampleModel);
                break;
            }
            default: {
                this.compressionType = 0;
                break;
            }
        }
        if (!this.canEncodeImage(this.compressionType, colorModel, sampleModel)) {
            throw new IOException("Image can not be encoded with compression type " + BMPCompressionTypes.getName(this.compressionType));
        }
        byte[] r2 = null;
        byte[] g = null;
        byte[] b = null;
        byte[] a = null;
        if (this.compressionType == 3) {
            n = DataBuffer.getDataTypeSize(sampleModel.getDataType());
            if (n != 16 && n != 32) {
                n = 32;
                n8 = 0;
            }
            n9 = this.w * n + 7 >> 3;
            n2 = 1;
            mapSize = 3;
            r2 = new byte[mapSize];
            g = new byte[mapSize];
            b = new byte[mapSize];
            a = new byte[mapSize];
            int redMask = 16711680;
            int greenMask = 65280;
            int blueMask = 255;
            if (n == 16) {
                if ((colorModel instanceof DirectColorModel)) {
                    DirectColorModel directColorModel = (DirectColorModel)colorModel;
                    redMask = directColorModel.getRedMask();
                    greenMask = directColorModel.getGreenMask();
                    blueMask = directColorModel.getBlueMask();
                }
                else {
                    throw new IOException("Image can not be encoded with compression type " + BMPCompressionTypes.getName(this.compressionType));
                }
            }
            this.writeMaskToPalette(redMask, 0, r2, g, b, a);
            this.writeMaskToPalette(greenMask, 1, r2, g, b, a);
            this.writeMaskToPalette(blueMask, 2, r2, g, b, a);
            if (n8 == 0) {
                (this.bitMasks = new int[3])[0] = redMask;
                this.bitMasks[1] = greenMask;
                this.bitMasks[2] = blueMask;
                (this.bitPos = new int[3])[0] = this.firstLowBit(redMask);
                this.bitPos[1] = this.firstLowBit(greenMask);
                this.bitPos[2] = this.firstLowBit(blueMask);
            }
            if ((colorModel instanceof IndexColorModel)) {
                indexColorModel = (IndexColorModel)colorModel;
            }
        }
        else if ((colorModel instanceof IndexColorModel)) {
            n2 = 1;
            indexColorModel = (IndexColorModel)colorModel;
            mapSize = indexColorModel.getMapSize();
            if (mapSize <= 2) {
                n = 1;
                n9 = this.w + 7 >> 3;
            }
            else if (mapSize <= 16) {
                n = 4;
                n9 = this.w + 1 >> 1;
            }
            else if (mapSize <= 256) {
                n = 8;
            }
            else {
                n = 24;
                n2 = 0;
                mapSize = 0;
                n9 = this.w * 3;
            }
            if (n2 == 1) {
                r2 = new byte[mapSize];
                g = new byte[mapSize];
                b = new byte[mapSize];
                a = new byte[mapSize];
                indexColorModel.getAlphas(a);
                indexColorModel.getReds(r2);
                indexColorModel.getGreens(g);
                indexColorModel.getBlues(b);
            }
        }
        else if (n6 == 1) {
            n2 = 1;
            mapSize = 256;
            n = sampleSize[0];
            n9 = this.w * n + 7 >> 3;
            r2 = new byte[256];
            g = new byte[256];
            b = new byte[256];
            a = new byte[256];
            for (int n12 = 0; n12 < 256; ++n12) {
                r2[n12] = (byte)n12;
                g[n12] = (byte)n12;
                b[n12] = (byte)n12;
                a[n12] = -1;
            }
        }
        else if (sampleModel instanceof SinglePixelPackedSampleModel && n5) {
            int[] sampleSize2 = sampleModel.getSampleSize();
            int n10 = 0;
            for (int length = sampleSize2.length, n11 = 0; n11 < length; ++n11) {
                n10 += sampleSize2[n11];
            }
            n = this.roundBpp(n10);
            if (n != DataBuffer.getDataTypeSize(sampleModel.getDataType())) {
                n8 = 0;
            }
            n9 = this.w * n + 7 >> 3;
        }
        int n13 = 0;
        int n14 = 0;
        int n15 = 0;
        int n16 = n9 % 4;
        if (n16 != 0) {
            n16 = 4 - n16;
        }
        int n17 = 54 + mapSize * 4;
        int n18 = (n9 + n16) * this.h;
        int n19 = n18 + n17;
        int n20 = 40;
        long streamPosition = this.stream.getStreamPosition();
        this.writeFileHeader(n19, n17);
        if (this.compressionType == 0 || this.compressionType == 3) {
            this.isTopDown = bmpImageWriteParam.isTopDown();
        }
        else {
            this.isTopDown = false;
        }
        this.writeInfoHeader(n20, n);
        this.stream.writeInt(this.compressionType);
        this.stream.writeInt(n18);
        this.stream.writeInt(n13);
        this.stream.writeInt(n14);
        this.stream.writeInt(n15);
        this.stream.writeInt(mapSize);
        if (n2 == 1) {
            if (this.compressionType == 3) {
                for (int n22 = 0; n22 < 3; ++n22) {
                    this.stream.writeInt((a[n22] & 0xFF) + (r2[n22] & 0xFF) * 256 + (g[n22] & 0xFF) * 65536 + (b[n22] & 0xFF) * 16777216);
                }
            }
            else {
                for (int n21 = 0; n21 < mapSize; ++n21) {
                    this.stream.writeByte((int)b[n21]);
                    this.stream.writeByte((int)g[n21]);
                    this.stream.writeByte((int)r2[n21]);
                    this.stream.writeByte((int)a[n21]);
                }
            }
        }
        int n23 = this.w * n6;
        int[] iArray = new int[n23 * sourceXSubsampling];
        this.bpixels = new byte[n9];
        if (this.compressionType == 4 || this.compressionType == 5) {
            this.embedded_stream = new ByteArrayOutputStream();
            this.writeEmbedded(iioImage, bmpImageWriteParam);
            this.embedded_stream.flush();
            int size = this.embedded_stream.size();
            long streamPosition3 = this.stream.getStreamPosition();
            int n38 = n17 + size;
            this.stream.seek(streamPosition);
            this.writeSize(n38, 2);
            this.stream.seek(streamPosition);
            this.writeSize(size, 34);
            this.stream.seek(streamPosition3);
            this.stream.write(this.embedded_stream.toByteArray());
            this.embedded_stream = null;
            if (this.abortRequested()) {
                this.processWriteAborted();
            }
            else {
                this.processImageComplete();
                this.stream.flushBefore(this.stream.getStreamPosition());
            }
            return;
        }
        else {
            int n24 = bandOffsets[0];
            for (int n25 = 1; n25 < bandOffsets.length; ++n25) {
                if (bandOffsets[n25] > n24) {
                    n24 = bandOffsets[n25];
                }
            }
            int[] array = new int[n24 + 1];
            if (n8 != 0 && n5) {
                n9 /= DataBuffer.getDataTypeSize(dataType) >> 3;
            }
            for (int n26 = 0; n26 < this.h && !this.abortRequested(); ++n26) {
                int n27 = y + n26;
                if (!this.isTopDown) {
                    n27 = y + this.h - n26 - 1;
                }
                Rectangle o = new Rectangle(x * sourceXSubsampling + n3, n27 * sourceYSubsampling + n4, (this.w - 1) * sourceXSubsampling + 1, 1);
                if (hasRaster == 0) {
                    raster = renderedImage.getData(o);
                }
                if (n8 != 0 && n5) {
                    SampleModel sampleModel2 = raster.getSampleModel();
                    int n32 = 0;
                    int x2 = o.x - raster.getSampleModelTranslateX();
                    int y2 = o.y - raster.getSampleModelTranslateY();
                    if ((sampleModel2 instanceof ComponentSampleModel)) {
                        ComponentSampleModel componentSampleModel = (ComponentSampleModel)sampleModel2;
                        n32 = componentSampleModel.getOffset(x2, y2, 0);
                        for (int n33 = 1; n33 < componentSampleModel.getNumBands(); ++n33) {
                            if (n32 > componentSampleModel.getOffset(x2, y2, n33)) {
                                n32 = componentSampleModel.getOffset(x2, y2, n33);
                            }
                        }
                    }
                    else if ((sampleModel2 instanceof MultiPixelPackedSampleModel)) {
                        n32 = ((MultiPixelPackedSampleModel)sampleModel2).getOffset(x2, y2);
                    }
                    else if ((sampleModel2 instanceof SinglePixelPackedSampleModel)) {
                        n32 = ((SinglePixelPackedSampleModel)sampleModel2).getOffset(x2, y2);
                    }
                    if (this.compressionType == 0 || this.compressionType == 3) {
                        switch (dataType) {
                            case 0: {
                                this.stream.write(((DataBufferByte)raster.getDataBuffer()).getData(), n32, n9);
                                break;
                            }
                            case 2: {
                                this.stream.writeShorts(((DataBufferShort)raster.getDataBuffer()).getData(), n32, n9);
                                break;
                            }
                            case 1: {
                                this.stream.writeShorts(((DataBufferUShort)raster.getDataBuffer()).getData(), n32, n9);
                                break;
                            }
                            case 3: {
                                this.stream.writeInts(((DataBufferInt)raster.getDataBuffer()).getData(), n32, n9);
                                break;
                            }
                        }
                        for (int n36 = 0; n36 < n16; ++n36) {
                            this.stream.writeByte(0);
                        }
                    }
                    else if (this.compressionType == 2) {
                        if (this.bpixels == null || this.bpixels.length < n23) {
                            this.bpixels = new byte[n23];
                        }
                        raster.getPixels(o.x, o.y, o.width, o.height, iArray);
                        for (int n35 = 0; n35 < n23; ++n35) {
                            this.bpixels[n35] = (byte)iArray[n35];
                        }
                        this.encodeRLE4(this.bpixels, n23);
                    }
                    else if (this.compressionType == 1) {
                        if (this.bpixels == null || this.bpixels.length < n23) {
                            this.bpixels = new byte[n23];
                        }
                        raster.getPixels(o.x, o.y, o.width, o.height, iArray);
                        for (int n34 = 0; n34 < n23; ++n34) {
                            this.bpixels[n34] = (byte)iArray[n34];
                        }
                        this.encodeRLE8(this.bpixels, n23);
                    }
                }
                else {
                    raster.getPixels(o.x, o.y, o.width, o.height, iArray);
                    if (sourceXSubsampling != 1 || n24 != n6 - 1) {
                        for (int n28 = 0, n29 = 0, n30 = 0; n28 < this.w; ++n28, n29 += sourceXSubsampling * n6, n30 += n6) {
                            System.arraycopy(iArray, n29, array, 0, array.length);
                            for (int n31 = 0; n31 < n6; ++n31) {
                                iArray[n30 + n31] = array[sourceBands[n31]];
                            }
                        }
                    }
                    this.writePixels(0, n23, n, iArray, n16, n6, indexColorModel);
                }
                this.processImageProgress(100.0f * ((float)n26 / (float)this.h));
            }
            if (this.compressionType == 2 || this.compressionType == 1) {
                this.stream.writeByte(0);
                this.stream.writeByte(1);
                this.incCompImageSize(2);
                int compImageSize = this.compImageSize;
                int n37 = this.compImageSize + n17;
                long streamPosition2 = this.stream.getStreamPosition();
                this.stream.seek(streamPosition);
                this.writeSize(n37, 2);
                this.stream.seek(streamPosition);
                this.writeSize(compImageSize, 34);
                this.stream.seek(streamPosition2);
            }
            if (this.abortRequested()) {
                this.processWriteAborted();
            }
            else {
                this.processImageComplete();
                this.stream.flushBefore(this.stream.getStreamPosition());
            }
            return;
        }
    }
    
    private void writePixels(int n, int n2, int n3, int[] array, int n4, int n5, IndexColorModel indexColorModel) throws IOException {
        int n6 = 0;
        switch (n3) {
            case 1: {
                for (int i = 0; i < n2 / 8; ++i) {
                    this.bpixels[n6++] = (byte)(array[n++] << 7 | array[n++] << 6 | array[n++] << 5 | array[n++] << 4 | array[n++] << 3 | array[n++] << 2 | array[n++] << 1 | array[n++]);
                }
                if (n2 % 8 > 0) {
                    int n7 = 0;
                    for (int j = 0; j < n2 % 8; ++j) {
                        n7 |= array[n++] << 7 - j;
                    }
                    this.bpixels[n6++] = (byte)n7;
                }
                this.stream.write(this.bpixels, 0, (n2 + 7) / 8);
                break;
            }
            case 4: {
                if (this.compressionType == 2) {
                    byte[] array2 = new byte[n2];
                    for (int l = 0; l < n2; ++l) {
                        array2[l] = (byte)array[n++];
                    }
                    this.encodeRLE4(array2, n2);
                    break;
                }
                for (int k = 0; k < n2 / 2; ++k) {
                    this.bpixels[n6++] = (byte)(array[n++] << 4 | array[n++]);
                }
                if (n2 % 2 == 1) {
                    this.bpixels[n6++] = (byte)(array[n] << 4);
                }
                this.stream.write(this.bpixels, 0, (n2 + 1) / 2);
                break;
            }
            case 8: {
                if (this.compressionType == 1) {
                    for (int n9 = 0; n9 < n2; ++n9) {
                        this.bpixels[n9] = (byte)array[n++];
                    }
                    this.encodeRLE8(this.bpixels, n2);
                    break;
                }
                for (int n8 = 0; n8 < n2; ++n8) {
                    this.bpixels[n8] = (byte)array[n++];
                }
                this.stream.write(this.bpixels, 0, n2);
                break;
            }
            case 16: {
                if (this.spixels == null) {
                    this.spixels = new short[n2 / n5];
                }
                for (int n10 = 0, n11 = 0; n10 < n2; ++n11) {
                    this.spixels[n11] = 0;
                    if (this.compressionType == 0) {
                        this.spixels[n11] = (short)((0x1F & array[n10]) << 10 | (0x1F & array[n10 + 1]) << 5 | (0x1F & array[n10 + 2]));
                        n10 += 3;
                    }
                    else {
                        for (int n12 = 0; n12 < n5; ++n12, ++n10) {
                            short[] spixels = this.spixels;
                            int n13 = n11;
                            spixels[n13] |= (array[n10] << this.bitPos[n12] & this.bitMasks[n12]);
                        }
                    }
                }
                this.stream.writeShorts(this.spixels, 0, this.spixels.length);
                break;
            }
            case 24: {
                if (n5 == 3) {
                    for (int n16 = 0; n16 < n2; n16 += 3) {
                        this.bpixels[n6++] = (byte)array[n + 2];
                        this.bpixels[n6++] = (byte)array[n + 1];
                        this.bpixels[n6++] = (byte)array[n];
                        n += 3;
                    }
                    this.stream.write(this.bpixels, 0, n2);
                    break;
                }
                int mapSize = indexColorModel.getMapSize();
                byte[] r = new byte[mapSize];
                byte[] g = new byte[mapSize];
                byte[] b = new byte[mapSize];
                indexColorModel.getReds(r);
                indexColorModel.getGreens(g);
                indexColorModel.getBlues(b);
                for (int n14 = 0; n14 < n2; ++n14) {
                    int n15 = array[n];
                    this.bpixels[n6++] = b[n15];
                    this.bpixels[n6++] = g[n15];
                    this.bpixels[n6++] = b[n15];
                    ++n;
                }
                this.stream.write(this.bpixels, 0, n2 * 3);
                break;
            }
            case 32: {
                if (this.ipixels == null) {
                    this.ipixels = new int[n2 / n5];
                }
                if (n5 == 3) {
                    for (int n18 = 0, n19 = 0; n18 < n2; ++n19) {
                        this.ipixels[n19] = 0;
                        if (this.compressionType == 0) {
                            this.ipixels[n19] = ((0xFF & array[n18 + 2]) << 16 | (0xFF & array[n18 + 1]) << 8 | (0xFF & array[n18]));
                            n18 += 3;
                        }
                        else {
                            for (int n20 = 0; n20 < n5; ++n20, ++n18) {
                                int[] ipixels = this.ipixels;
                                int n21 = n19;
                                ipixels[n21] |= (array[n18] << this.bitPos[n20] & this.bitMasks[n20]);
                            }
                        }
                    }
                }
                else {
                    for (int n17 = 0; n17 < n2; ++n17) {
                        if (indexColorModel != null) {
                            this.ipixels[n17] = indexColorModel.getRGB(array[n17]);
                        }
                        else {
                            this.ipixels[n17] = (array[n17] << 16 | array[n17] << 8 | array[n17]);
                        }
                    }
                }
                this.stream.writeInts(this.ipixels, 0, this.ipixels.length);
                break;
            }
        }
        if (this.compressionType == 0 || this.compressionType == 3) {
            for (int n22 = 0; n22 < n4; ++n22) {
                this.stream.writeByte(0);
            }
        }
    }
    
    private void encodeRLE8(byte[] array, int n) throws IOException {
        int n2 = 1;
        int n3 = -1;
        int i = -1;
        int n4 = array[++i];
        byte[] array2 = new byte[256];
        while (i < n - 1) {
            int n5 = array[++i];
            if (n5 == n4) {
                if (n3 >= 3) {
                    this.stream.writeByte(0);
                    this.stream.writeByte(n3);
                    this.incCompImageSize(2);
                    for (int l = 0; l < n3; ++l) {
                        this.stream.writeByte((int)array2[l]);
                        this.incCompImageSize(1);
                    }
                    if (!this.isEven(n3)) {
                        this.stream.writeByte(0);
                        this.incCompImageSize(1);
                    }
                }
                else if (n3 > -1) {
                    for (int k = 0; k < n3; ++k) {
                        this.stream.writeByte(1);
                        this.stream.writeByte((int)array2[k]);
                        this.incCompImageSize(2);
                    }
                }
                n3 = -1;
                if (++n2 == 256) {
                    this.stream.writeByte(n2 - 1);
                    this.stream.writeByte(n4);
                    this.incCompImageSize(2);
                    n2 = 1;
                }
            }
            else {
                if (n2 > 1) {
                    this.stream.writeByte(n2);
                    this.stream.writeByte(n4);
                    this.incCompImageSize(2);
                }
                else if (n3 < 0) {
                    array2[++n3] = n4;
                    array2[++n3] = n5;
                }
                else if (n3 < 254) {
                    array2[++n3] = n5;
                }
                else {
                    this.stream.writeByte(0);
                    this.stream.writeByte(n3 + 1);
                    this.incCompImageSize(2);
                    for (int j = 0; j <= n3; ++j) {
                        this.stream.writeByte((int)array2[j]);
                        this.incCompImageSize(1);
                    }
                    this.stream.writeByte(0);
                    this.incCompImageSize(1);
                    n3 = -1;
                }
                n4 = n5;
                n2 = 1;
            }
            if (i != n - 1) {
                continue;
            }
            if (n3 == -1) {
                this.stream.writeByte(n2);
                this.stream.writeByte(n4);
                this.incCompImageSize(2);
                n2 = 1;
            }
            else if (n3 >= 2) {
                this.stream.writeByte(0);
                this.stream.writeByte(n3 + 1);
                this.incCompImageSize(2);
                for (int n7 = 0; n7 <= n3; ++n7) {
                    this.stream.writeByte((int)array2[n7]);
                    this.incCompImageSize(1);
                }
                if (!this.isEven(n3 + 1)) {
                    this.stream.writeByte(0);
                    this.incCompImageSize(1);
                }
            }
            else if (n3 > -1) {
                for (int n6 = 0; n6 <= n3; ++n6) {
                    this.stream.writeByte(1);
                    this.stream.writeByte((int)array2[n6]);
                    this.incCompImageSize(2);
                }
            }
            this.stream.writeByte(0);
            this.stream.writeByte(0);
            this.incCompImageSize(2);
        }
    }
    
    private void encodeRLE4(byte[] array, int n) throws IOException {
        int n2 = 2;
        int n3 = -1;
        int i = -1;
        byte[] array2 = new byte[256];
        int n4 = array[++i];
        int n5 = array[++i];
        while (i < n - 2) {
            int n6 = array[++i];
            int n7 = array[++i];
            if (n6 == n4) {
                if (n3 >= 4) {
                    this.stream.writeByte(0);
                    this.stream.writeByte(n3 - 1);
                    this.incCompImageSize(2);
                    for (int k = 0; k < n3 - 2; k += 2) {
                        this.stream.writeByte((int)(byte)(array2[k] << 4 | array2[k + 1]));
                        this.incCompImageSize(1);
                    }
                    if (!this.isEven(n3 - 1)) {
                        this.stream.writeByte(array2[n3 - 2] << 4 | 0x0);
                        this.incCompImageSize(1);
                    }
                    if (!this.isEven((int)Math.ceil((double)((n3 - 1) / 2)))) {
                        this.stream.writeByte(0);
                        this.incCompImageSize(1);
                    }
                }
                else if (n3 > -1) {
                    this.stream.writeByte(2);
                    this.stream.writeByte(array2[0] << 4 | array2[1]);
                    this.incCompImageSize(2);
                }
                n3 = -1;
                if (n7 == n5) {
                    n2 += 2;
                    if (n2 == 256) {
                        this.stream.writeByte(n2 - 1);
                        this.stream.writeByte(n4 << 4 | n5);
                        this.incCompImageSize(2);
                        n2 = 2;
                        if (i < n - 1) {
                            n4 = n5;
                            n5 = array[++i];
                        }
                        else {
                            this.stream.writeByte(1);
                            this.stream.writeByte(n5 << 4 | 0x0);
                            this.incCompImageSize(2);
                            n2 = -1;
                        }
                    }
                }
                else {
                    ++n2;
                    int n9 = n4 << 4 | n5;
                    this.stream.writeByte(n2);
                    this.stream.writeByte(n9);
                    this.incCompImageSize(2);
                    n2 = 2;
                    n4 = n7;
                    if (i < n - 1) {
                        n5 = array[++i];
                    }
                    else {
                        this.stream.writeByte(1);
                        this.stream.writeByte(n7 << 4 | 0x0);
                        this.incCompImageSize(2);
                        n2 = -1;
                    }
                }
            }
            else {
                if (n2 > 2) {
                    int n8 = n4 << 4 | n5;
                    this.stream.writeByte(n2);
                    this.stream.writeByte(n8);
                    this.incCompImageSize(2);
                }
                else if (n3 < 0) {
                    array2[++n3] = n4;
                    array2[++n3] = n5;
                    array2[++n3] = n6;
                    array2[++n3] = n7;
                }
                else if (n3 < 253) {
                    array2[++n3] = n6;
                    array2[++n3] = n7;
                }
                else {
                    this.stream.writeByte(0);
                    this.stream.writeByte(n3 + 1);
                    this.incCompImageSize(2);
                    for (int j = 0; j < n3; j += 2) {
                        this.stream.writeByte((int)(byte)(array2[j] << 4 | array2[j + 1]));
                        this.incCompImageSize(1);
                    }
                    this.stream.writeByte(0);
                    this.incCompImageSize(1);
                    n3 = -1;
                }
                n4 = n6;
                n5 = n7;
                n2 = 2;
            }
            if (i < n - 2) {
                continue;
            }
            if (n3 == -1 && n2 >= 2) {
                if (i == n - 2) {
                    if (array[++i] == n4) {
                        ++n2;
                        int n12 = n4 << 4 | n5;
                        this.stream.writeByte(n2);
                        this.stream.writeByte(n12);
                        this.incCompImageSize(2);
                    }
                    else {
                        int n10 = n4 << 4 | n5;
                        this.stream.writeByte(n2);
                        this.stream.writeByte(n10);
                        this.stream.writeByte(1);
                        this.stream.writeByte(array[i] << 4 | 0x0);
                        int n11 = array[i] << 4 | 0x0;
                        this.incCompImageSize(4);
                    }
                }
                else {
                    this.stream.writeByte(n2);
                    this.stream.writeByte(n4 << 4 | n5);
                    this.incCompImageSize(2);
                }
            }
            else if (n3 > -1) {
                if (i == n - 2) {
                    array2[++n3] = array[++i];
                }
                if (n3 >= 2) {
                    this.stream.writeByte(0);
                    this.stream.writeByte(n3 + 1);
                    this.incCompImageSize(2);
                    for (int l = 0; l < n3; l += 2) {
                        this.stream.writeByte((int)(byte)(array2[l] << 4 | array2[l + 1]));
                        this.incCompImageSize(1);
                    }
                    if (!this.isEven(n3 + 1)) {
                        this.stream.writeByte(array2[n3] << 4 | 0x0);
                        this.incCompImageSize(1);
                    }
                    if (!this.isEven((int)Math.ceil((double)((n3 + 1) / 2)))) {
                        this.stream.writeByte(0);
                        this.incCompImageSize(1);
                    }
                }
                else {
                    switch (n3) {
                        case 0: {
                            this.stream.writeByte(1);
                            this.stream.writeByte(array2[0] << 4 | 0x0);
                            this.incCompImageSize(2);
                            break;
                        }
                        case 1: {
                            this.stream.writeByte(2);
                            this.stream.writeByte(array2[0] << 4 | array2[1]);
                            this.incCompImageSize(2);
                            break;
                        }
                    }
                }
            }
            this.stream.writeByte(0);
            this.stream.writeByte(0);
            this.incCompImageSize(2);
        }
    }
    
    private synchronized void incCompImageSize(int n) {
        this.compImageSize += n;
    }
    
    private boolean isEven(int n) {
        return n % 2 == 0;
    }
    
    private void writeFileHeader(int n, int n2) throws IOException {
        this.stream.writeByte(66);
        this.stream.writeByte(77);
        this.stream.writeInt(n);
        this.stream.writeInt(0);
        this.stream.writeInt(n2);
    }
    
    private void writeInfoHeader(int n, int n2) throws IOException {
        this.stream.writeInt(n);
        this.stream.writeInt(this.w);
        this.stream.writeInt(this.isTopDown ? (-this.h) : this.h);
        this.stream.writeShort(1);
        this.stream.writeShort(n2);
    }
    
    private void writeSize(int n, int n2) throws IOException {
        this.stream.skipBytes(n2);
        this.stream.writeInt(n);
    }
    
    @Override
    public void reset() {
        super.reset();
        this.stream = null;
    }
    
    private void writeEmbedded(IIOImage iioImage, ImageWriteParam imageWriteParam) throws IOException {
        String s = (this.compressionType == 4) ? "jpeg" : "png";
        Iterator<ImageWriter> imageWritersByFormatName = ImageIO.getImageWritersByFormatName(s);
        ImageWriter imageWriter = null;
        if (imageWritersByFormatName.hasNext()) {
            imageWriter = (ImageWriter)imageWritersByFormatName.next();
        }
        if (imageWriter != null) {
            if (this.embedded_stream == null) {
                throw new RuntimeException("No stream for writing embedded image!");
            }
            imageWriter.addIIOWriteProgressListener(new BMPImageWriter.IIOWriteProgressAdapter() {
                @Override
                public void imageProgress(ImageWriter imageWriter, float n) {
                    BMPImageWriter.access$100(BMPImageWriter.this, n);
                }
            });
            imageWriter.addIIOWriteWarningListener(new IIOWriteWarningListener() {
                @Override
                public void warningOccurred(ImageWriter imageWriter, int n, String s) {
                    BMPImageWriter.access$200(BMPImageWriter.this, n, s);
                }
            });
            imageWriter.setOutput(ImageIO.createImageOutputStream(this.embedded_stream));
            ImageWriteParam defaultWriteParam = imageWriter.getDefaultWriteParam();
            defaultWriteParam.setDestinationOffset(imageWriteParam.getDestinationOffset());
            defaultWriteParam.setSourceBands(imageWriteParam.getSourceBands());
            defaultWriteParam.setSourceRegion(imageWriteParam.getSourceRegion());
            defaultWriteParam.setSourceSubsampling(imageWriteParam.getSourceXSubsampling(), imageWriteParam.getSourceYSubsampling(), imageWriteParam.getSubsamplingXOffset(), imageWriteParam.getSubsamplingYOffset());
            imageWriter.write(null, iioImage, defaultWriteParam);
            return;
        }
        else {
            throw new RuntimeException(I18N.getString("BMPImageWrite5") + " " + s);
        }
    }
    
    private int firstLowBit(int n) {
        int n2 = 0;
        while ((n & 0x1) == 0x0) {
            ++n2;
            n >>>= 1;
        }
        return n2;
    }
    
    protected int getPreferredCompressionType(ColorModel colorModel, SampleModel sampleModel) {
        return this.getPreferredCompressionType(new ImageTypeSpecifier(colorModel, sampleModel));
    }
    
    protected int getPreferredCompressionType(ImageTypeSpecifier imageTypeSpecifier) {
        if (imageTypeSpecifier.getBufferedImageType() == 8) {
            return 3;
        }
        return 0;
    }
    
    protected boolean canEncodeImage(int n, ColorModel colorModel, SampleModel sampleModel) {
        return this.canEncodeImage(n, new ImageTypeSpecifier(colorModel, sampleModel));
    }
    
    protected boolean canEncodeImage(int n, ImageTypeSpecifier imageTypeSpecifier) {
        if (!this.getOriginatingProvider().canEncodeImage(imageTypeSpecifier)) {
            return false;
        }
        imageTypeSpecifier.getBufferedImageType();
        int pixelSize = imageTypeSpecifier.getColorModel().getPixelSize();
        if (this.compressionType == 2 && pixelSize != 4) {
            return false;
        }
        if (this.compressionType == 1 && pixelSize != 8) {
            return false;
        }
        if (pixelSize == 16) {
            boolean n2 = false;
            boolean n3 = false;
            SampleModel sampleModel = imageTypeSpecifier.getSampleModel();
            if ((sampleModel instanceof SinglePixelPackedSampleModel)) {
                int[] sampleSize = ((SinglePixelPackedSampleModel)sampleModel).getSampleSize();
                n2 = true;
                n3 = true;
                for (int i = 0; i < sampleSize.length; ++i) {
                    n2 &= (sampleSize[i] == 5);
                    n3 &= (sampleSize[i] == 5 || (i == 1 && sampleSize[i] == 6));
                }
            }
            return (this.compressionType == 0 && n2) || (this.compressionType == 3 && n3);
        }
        else {
            return true;
        }
    }
    
    protected void writeMaskToPalette(int n, int n2, byte[] array, byte[] array2, byte[] array3, byte[] array4) {
        array3[n2] = (byte)(0xFF & n >> 24);
        array2[n2] = (byte)(0xFF & n >> 16);
        array[n2] = (byte)(0xFF & n >> 8);
        array4[n2] = (byte)(0xFF & n);
    }
    
    private int roundBpp(int n) {
        if (n <= 8) {
            return 8;
        }
        if (n <= 16) {
            return 16;
        }
        if (n <= 24) {
            return 24;
        }
        return 32;
    }
    
    static void access$100(BMPImageWriter bmpImageWriter, float percentageDone) {
        bmpImageWriter.processImageProgress(percentageDone);
    }
    
    static void access$200(BMPImageWriter bmpImageWriter, int imageIndex, String warning) {
        bmpImageWriter.processWarningOccurred(imageIndex, warning);
    }
    
    private class IIOWriteProgressAdapter implements IIOWriteProgressListener
    {
        @Override
        public void imageComplete(ImageWriter imageWriter) {
        }
        
        @Override
        public void imageProgress(ImageWriter imageWriter, float n) {
        }
        
        @Override
        public void imageStarted(ImageWriter imageWriter, int n) {
        }
        
        @Override
        public void thumbnailComplete(ImageWriter imageWriter) {
        }
        
        @Override
        public void thumbnailProgress(ImageWriter imageWriter, float n) {
        }
        
        @Override
        public void thumbnailStarted(ImageWriter imageWriter, int n, int n2) {
        }
        
        @Override
        public void writeAborted(ImageWriter imageWriter) {
        }
    }
}
